/*

    File: LODMethod.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>
#include <gl/gl.h>
#include <math.h>
#include <assert.h>

#include "../m_system3d/main.h"
#include "../m_foundation/main.h"
#include "../m_foundation/const.h"

#include "landscape.h"
#include "lodmethod.h"

bool LODMethod::init(int width, int height, ls_vertex_t *data) {return true;}
void LODMethod::render(int width, int height, ls_vertex_t *data) {}
void LODMethod::shutdown(int width, int height, ls_vertex_t *data) {}
const char *LODMethod::desc(void) {return "LOD superclass";}

//256x256 tristrip mesh
//vertex data float
//cullface enable
//depth test enabled
//texture2d disabled
//27FPS
//
//color data unsigned byte 4
//17FPS
//
//color data float 3
//12FPS
//
//assertion:
//	the FPS is going down 'cuz the size is going up.  simple as that
//	the number of components or the format do not noticeably affect performance
//
//no color data
//texture data generated
//14FPS
//
//vertex data
//color data ub4
//texture data gen'd
//10FPS
//
//texgen on
//11FPS
//
//texture data none
//16FPS
//
//color data none
//19FPS
//
void LODMethodBrute::render(int width, int height, ls_vertex_t *data) {

	for (int y = 0; y < height-1; y++) {
		glBegin(GL_TRIANGLE_STRIP);
		for (int x = 0; x < width; x++, data++) {

			data += width;
			y++;
#ifndef LIGHTMAP_TEXTURE
			glColor3ubv((unsigned char *)&data->color);
#endif
			glVertex3f((float)x, (float)y, data->height);

			data -= width;
			y--;
#ifndef LIGHTMAP_TEXTURE
			glColor3ubv((unsigned char *)&data->color);
#endif
			glVertex3f((float)x, (float)y, data->height);
		}
		glEnd();
	}
}

const char *LODMethodBrute::desc(void) { return "brute force"; }

LODMethodTrial1::LODMethodTrial1() {

	VectorSet(&stepCoeff, 1, 0.02f, 0);
	maxQuadsPerBlock = 48;
	forceAlignPowerOfTwo = false;

	new CVarw(&stepCoeff, "t1_stepCoeff");
	new CVarw(&maxQuadsPerBlock, "t1_maxQuadsPerBlock");
	new CVarw(&forceAlignPowerOfTwo, "t1_forceAlignPowerOfTwo");
}

void LODMethodTrial1::render(int width, int height, ls_vertex_t *base) {

	int tileWidth = width - 1;
	int tileHeight = height - 1;

	int quadsPerBlock = (width + height) / 16;	//32;
	if (quadsPerBlock > maxQuadsPerBlock) {
		quadsPerBlock = maxQuadsPerBlock;
	}

	int blocksWide = tileWidth / quadsPerBlock;
	int blocksHigh = tileHeight / quadsPerBlock;

	for (int j = 0; j < blocksHigh; j++) {
		for (int i = 0; i < blocksWide; i++) {

			int startX = (i * tileWidth) / blocksWide;
			int startY = (j * tileHeight) / blocksHigh;
			int endX = ((i + 1) * tileWidth) / blocksWide;
			int endY = ((j + 1) * tileHeight) / blocksHigh;

			int cx = (startX + endX) >> 1;
			int cy = (startY + endY) >> 1;
			float cz = base[cx + cy * width].height;
			vector_t del = {
				(float)cx - main_camera->position.x,
				(float)cy - main_camera->position.y,
				cz - main_camera->position.z};
			float dist = (float)sqrt(del.x * del.x + del.y * del.y + del.z * del.z);
			int step = (int)(stepCoeff.x + stepCoeff.y * dist + stepCoeff.z * dist * dist);
			if (forceAlignPowerOfTwo) {
				//copy of the fastRoundDownPowerOfTwo code:
				step >>= 1;
				for (int newStep=1; step; step >>= 1, newStep <<= 1);
				step = newStep;
			}
			if (step < 1) step = 1;
			
			int xStep = step;
			int yStep = step;

//currently only implemented under render method #1
#define USE_SINGLE_TRISTRIP

			if (startY < endY) {

#ifdef USE_SINGLE_TRISTRIP
				glBegin(GL_TRIANGLE_STRIP);
				glVertex3f((float)endX, (float)startY, base[endX + width * startY].height);
				bool xReverse = true;
#endif //USE_SINGLE_TRISTRIP

				int y = startY;
				bool doneY = false;
				do {
					int nextY = y + yStep;
					if (nextY >= endY) {
						nextY = endY;
						doneY = true;
					}

#ifndef USE_SINGLE_TRISTRIP
					glBegin(GL_TRIANGLE_STRIP);
#endif //USE_SINGLE_TRISTRIP

					if (startX < endX) {
#ifdef USE_SINGLE_TRISTRIP
						int x = xReverse ? endX : startX;
						xStep = xReverse ? -step : step;
#else //USE_SINGLE_TRISTRIP
						int x = startX;
#endif //USE_SINGLE_TRISTRIP
						bool doneX = false;
						do {

#ifdef USE_SINGLE_TRISTRIP

							int nextX = x + xStep;
							if (xReverse) {
								if (nextX <= startX) {
									nextX = startX;
									doneX = true;
								}
							} else {
								if (nextX >= endX) {
									nextX = endX;
									doneX = true;
								}
							}
							glVertex3f((float)x, (float)nextY, base[x + nextY * width].height);
							glVertex3f((float)nextX, (float)y, base[nextX + y * width].height);
							x = nextX;

#else //USE_SINGLE_TRISTRIP

							if (x >= endX) {	//make sure we hit the last one.
								x = endX;
								doneX = true;
							}

							ls_vertex_t *data = base + x + nextY * width;
#ifndef LIGHTMAP_TEXTURE
							glColor3ubv((unsigned char *)&data->color);
#endif
							glVertex3f((float)x, (float)nextY, data->height);

							data = base + x + y * width;
#ifndef LIGHTMAP_TEXTURE
							glColor3ubv((unsigned char *)&data->color);
#endif
							glVertex3f((float)x, (float)y, data->height);

							x += xStep;

#endif //USE_SINGLE_TRISTRIP

						} while (!doneX);

#ifdef USE_SINGLE_TRISTRIP
						glVertex3f((float)x, (float)nextY, base[x + nextY * width].height);
						xReverse = !xReverse;
#endif //USE_SINGLE_TRISTRIP

					}

#ifndef USE_SINGLE_TRISTRIP
					glEnd();
#endif //USE_SINGLE_TRISTRIP
					y = nextY;
				} while (!doneY);

#ifdef USE_SINGLE_TRISTRIP
				glEnd();
#endif //USE_SINGLE_TRISTRIP
			}
		}
	}
}

const char *LODMethodTrial1::desc(void) { return "trial 1 - solid step"; }

//lets divide the world up into 32x32 tiles
#define TRIAL2_QUADS_PER_BLOCK	32

LODMethodTrial2::LODMethodTrial2() {
	blockCount = 0;
	block = NULL;
}

bool LODMethodTrial2::isLeaf(int index) {
	return getFirstChild(index) >= blockCount;
}

//0 -> 1,2,3,4
//1 -> 5,6,7,8
//2 -> 9,10,11,12
//parent -> (child<<2)+{1,2,3,4}
//(child-1)>>2 -> parent
int LODMethodTrial2::getParent(int childIndex) {
	return (childIndex - 1) >> 2;
}

int LODMethodTrial2::getFirstChild(int parentIndex) {
	return (parentIndex << 2) + 1;
}

//gets the child index relative to the parent
//returns 0,1,2,3
int LODMethodTrial2::getChildRelative(int childIndex) {
	if (!childIndex) {
		//error
		return 0;
	}
	return (childIndex - 1) & 3;
}

void LODMethodTrial2::getBlockStats(int index, int *x0, int *y0, int *x1, int *y1) {
	if (!index) {
		*x0 = 0;
		*y0 = 0;
		*x1 = width-1;
		*y1 = height-1;
		return;
	}

	int parent = getParent(index);

	int cx0, cy0, cx1, cy1;
	getBlockStats(parent, &cx0, &cy0, &cx1, &cy1);
	int cxm = (cx0 + cx1) >> 1;
	int cym = (cy0 + cy1) >> 1;

	int rel = getChildRelative(index);
	
	*x0 = (rel & 1) ? cxm : cx0;
	*y0 = (rel & 2) ? cym : cy0;
	*x1 = (rel & 1) ? cx1 : cxm;
	*y1 = (rel & 2) ? cy1 : cym;
}

//assumes that everything's been created except vtxLERPData, which contains a 1-1 mapping for vtxData
bool LODMethodTrial2::LODReduce(ls_block_t *b) {

	int xmin, ymin, xmax, ymax;
	getBlockStats(b - block, &xmin, &ymin, &xmax, &ymax);

//	for (int j = b->vtxCount >> 2; j > 0; j--)
	for (;;) {

		{
			int collapseIndexCount = 0;
			//for all triangles
			for (int i = 0; i < b->indexCount; i+=3) {
				//if the triangle is not degenerate
				if (b->vtxLERPData[b->indexData[i+0]] == b->vtxLERPData[b->indexData[i+1]]
				 || b->vtxLERPData[b->indexData[i+1]] == b->vtxLERPData[b->indexData[i+2]]
				 || b->vtxLERPData[b->indexData[i+2]] == b->vtxLERPData[b->indexData[i+0]]
				) {
					continue;
				}
				collapseIndexCount += 3;
			}
			main_console->Print("index count: %d collapse index count: %d\n", b->indexCount, collapseIndexCount);
			if (collapseIndexCount <= b->indexCount >> 2) break;
		}


		int bestfrom = -1;
		int bestto = -1;
		float bestcost = fInf;

	//	//do the lod reduction (here ???)
	//	// ignore any vertices on patch corners
	//	// only reduce vertices on edges onto other vertices on edges
	//	// and ffa otherwise
	//	//
	//	//how does that look?
	//	//pass our function the index set and the vertex set
	//	//it will go through all the (valid) edges, compute the costs,
	//	//find the best cost edge, collapse it
		for (int i = 0; i < b->indexCount; i++) {
			int ii = ((i + 1) % 3) + (3 * (i / 3));	//next in the cycle

			int v1 = b->vtxLERPData[b->indexData[i]];
			int v2 = b->vtxLERPData[b->indexData[ii]];

			if (v1 == v2) continue;	//cant collapse an already-collapsed edge, can we =)
			//speaking of which
			//every time we collapse a vertex
			//we'll have to cycle through and find any other vertices already collapsed to it
			//and collapse them along with it

			int l1 = b->vtxData[v1];
			int l2 = b->vtxData[v2];
			assert(l1 != l2);
			//can we make this assertion?
			//assert that all vertices in a mesh are unique...
			//it should be taken care of elsewhere

			int x1 = l1 % width;
			int y1 = l1 / width;

			int x2 = l2 % width;
			int y2 = l2 / width;

#if 0	//allow edge collapses
			//1 = from, 2 = to (rhyming conveniantly)
			//we can't allow 'from' vertices to be on corners
			if ((x1 == xmin || x1 == xmax) && (y1 == ymin || y1 == ymax)) continue;
			if (x1 == xmin || x1 == xmax || y1 == ymin || y1 == ymax) {
				if (!(x2 == xmin || x2 == xmax || y2 == ymin || y2 == ymax)) {
					continue;
				}
			}
#else	//dont
			if (x1 == xmin || x1 == xmax || y1 == ymin || y1 == ymax) continue;
#endif

			//collapse cost of the edge . . .
			//dist right now =)
			float dx = (float)(x2 - x1);
			float dy = (float)(y2 - y1);
			float dz = data[l2].height - data[l1].height;
			float cost = dz * dz + dx * dx + dy * dy;

			if (cost < bestcost) {
				bestcost = cost;
				bestfrom = i;
				bestto = ii;
			}
		}

		if (bestfrom != -1) {
			assert(bestto != -1);
			assert(bestfrom != bestto);

			//bestfrom is the index from, referencing the vtxLERPData
			//bestto is the index to, also vtxLERPData

			//collapse

			//find targets
			int v1 = b->vtxLERPData[b->indexData[bestfrom]];
			int v2 = b->vtxLERPData[b->indexData[bestto]];

			//and reassign all - from the old, to the new
			for (int i = 0; i < b->vtxCount; i++) {
				if (b->vtxLERPData[i] == v1) {
					b->vtxLERPData[i] = v2;
				}
			}
		} else {
			main_console->Print("couldn't find any good collapse targets\n");
			return true;
		}
	}

	return true;
}

bool LODMethodTrial2::genBlockData(int index, int x0, int y0, int x1, int y1) {

	ls_block_t *b = block + index;

	//IF YOU'RE A LEAF THE BUILD YOUR OWN DATA SET
	int firstChild = getFirstChild(index);
	if (firstChild >= blockCount) {

		main_console->Print("generating leaf for block %d\n", index);

		int blockVtxWidth = (x1 - x0 + 1);
		int blockVtxHeight = (y1 - y0 + 1);
		b->vtxCount = blockVtxWidth * blockVtxHeight;
		b->vtxData = (int *)malloc(b->vtxCount * sizeof(int));
		if (!b->vtxData) {
			main_console->Print("failed to allocate block %d vertex data\n", index);
			return false;
		}

		//build vtxData
		int e,i,j;
		e = 0;
		for (j = 0; j < blockVtxHeight; j++) {
			for (i = 0; i < blockVtxWidth; i++) {
				b->vtxData[e] = (x0 + i) + (y0 + j) * width; e++;
			}
		}

		//if we're a leaf then gen the initial data.  3 vtx per tri, 2 tri per quad, WxH quad per block
		b->indexCount = (x1-x0)*(y1-y0)*6;
		b->indexData = (int *)malloc(b->indexCount * sizeof(int));
		if (!b->indexData) {
			main_console->Print("failed to allocate block %d index data\n", index);
			return false;
		}
		//build indexData
		e = 0;
		for (j = 0; j < blockVtxHeight-1; j++) {
			for (i = 0; i < blockVtxWidth-1; i++) {
				b->indexData[e] = i + j * blockVtxWidth;			e++;
				b->indexData[e] = (i+1) + j * blockVtxWidth;		e++;
				b->indexData[e] = (i+1) + (j+1) * blockVtxWidth;	e++;
				b->indexData[e] = (i+1) + (j+1) * blockVtxWidth;	e++;
				b->indexData[e] = i + (j+1) * blockVtxWidth;		e++;
				b->indexData[e] = i + j * blockVtxWidth;			e++;
			}
		}

	//IF NOT A LEAF THEN BUILD YOUR DATA SET FROM THE REDUCTION SET OF ALL YOUR CHILDREN
	} else {

		//first build the children
		int xm = (x0 + x1) >> 1;
		int ym = (y0 + y1) >> 1;
		//let c represent the child index relative to the parent {0-3}
		for (int c = 0; c < 4; c++) {
			int cx0 = (c & 1) ? xm : x0;
			int cy0 = (c & 2) ? ym : y0;
			int cx1 = (c & 1) ? x1 : xm;
			int cy1 = (c & 2) ? y1 : ym;
			if (!genBlockData(firstChild + c, cx0, cy0, cx1, cy1)) {
				main_console->Print("failed to allocate block %d child data\n", index);
				return false;
			}
		}

		//now construct this data off of them
		//count up and collect all non-degenerate triangles within all four child meshes

		//(this will all be repetative to the reduction code)

		//FIRST CONSTRUCT VERTEX DATA

		//cycle through all child vertices
		//for each of the LERP data elements,
		//if the index is unique (if none preceeding it match it)
		//then increment our new vertex count

		b->vtxCount = 0;

		//let c ref the child index in the block array
		for (c = firstChild; c < firstChild+4; c++) {
			ls_block_t *child = block + c;
			for (int i = 0; i < child->vtxCount; i++) {
				for (int j = 0; j < i; j++) {
					//if the lerp vertex is already found then 
					//its repetative - and dont count it
					if (child->vtxLERPData[i] == child->vtxLERPData[j]) {
						break;
					}
				}
				//if we didnt find any matching then the vertex is new so count it
				if (j == i) {
					b->vtxCount++;
				}
			}
		}

		b->vtxData = (int *)malloc(b->vtxCount * sizeof(int));
		if (!b->vtxData) {
			main_console->Print("failed to allocate block %d vertex LOD map\n", index);
			return false;
		}

		//let c ref the child index in the block array
		int e = 0;
		for (c = firstChild; c < firstChild+4; c++) {
			ls_block_t *child = block + c;
			for (int i = 0; i < child->vtxCount; i++) {
				for (int j = 0; j < i; j++) {
					//if the lerp vertex is already found then 
					//its repetative - and dont count it
					if (child->vtxLERPData[i] == child->vtxLERPData[j]) {
						break;
					}
				}
				//if we didnt find any matching then this vertex should be added
				if (j == i) {
					//get the vertex index relative to the whole mesh
					b->vtxData[e] = child->vtxData[child->vtxLERPData[i]];
					e++;
				}
			}
		}

		//SECOND CONSTRUCT TRIANGLE DATA

		b->indexCount = 0;

		//for all children
		for (c = firstChild; c < firstChild+4; c++) {
			ls_block_t *child = block + c;
			//for all triangles
			for (int i = 0; i < child->indexCount; i+=3) {
				//if the triangle is not degenerate
				if (child->vtxLERPData[child->indexData[i+0]] == child->vtxLERPData[child->indexData[i+1]]
				 || child->vtxLERPData[child->indexData[i+1]] == child->vtxLERPData[child->indexData[i+2]]
				 || child->vtxLERPData[child->indexData[i+2]] == child->vtxLERPData[child->indexData[i+0]]
				) {
					continue;
				}
				b->indexCount += 3;
			}
		}

		b->indexData = (int *)malloc(b->indexCount * sizeof(int));
		if (!b->indexData) {
			main_console->Print("failed to alloc block %d index data\n", index);
			return false;
		}

		//for all children
		e = 0;
		for (c = firstChild; c < firstChild+4; c++) {
			ls_block_t *child = block + c;
			//for all triangles
			for (int i = 0; i < child->indexCount; i+=3) {
				//if the triangle is not degenerate
				if (child->vtxLERPData[child->indexData[i+0]] == child->vtxLERPData[child->indexData[i+1]]
				 || child->vtxLERPData[child->indexData[i+1]] == child->vtxLERPData[child->indexData[i+2]]
				 || child->vtxLERPData[child->indexData[i+2]] == child->vtxLERPData[child->indexData[i+0]]
				) {
					continue;
				}
				for (int l = 0; l < 3; l++) {
					for (int k = 0; k < b->vtxCount; k++) {
						if (b->vtxData[k] == child->vtxData[child->vtxLERPData[child->indexData[i+l]]]) {
							b->indexData[e] = k;
							e++;
							break;
						}
					}
					if (k == b->vtxCount) {
						main_console->Print("after we got the vtx data, we couldnt match it up with the tri data\n");
						return false;
					}
				}
			}
		}
	}

	//FINALLY GENERATE YOUR OWN REDUCTION DATA

	//alloc our lerp table
	b->vtxLERPData = (int *)malloc(b->vtxCount * sizeof(int));
	if (!b->vtxLERPData) {
		main_console->Print("failed to alloc LERP data table\n");
		return false;
	}

	for (int i = 0; i < b->vtxCount; i++) {
			b->vtxLERPData[i] = i;
	}

	if (!LODReduce(b)) return false;

//	main_console->Print("created block %d range %d,%d to %d,%d\n", index, x0,y0, x1,y1);

	return true;
}

//LOD zero represents the full mesh.  recursively subdivide
bool LODMethodTrial2::init(int width, int height, ls_vertex_t *data) {
	this->width = width;
	this->height = height;
	this->data = data;

	//dim = log2(min(width,height)/alpha)
	int dim = height;
	if (dim > width) dim = width;
	dim /= TRIAL2_QUADS_PER_BLOCK;
	//calculate log2dim integer quickly
	for (int log2dim = -1; dim > 0; dim >>= 1, log2dim++);
	if (log2dim < 0) log2dim = 0;
	//sum up all 4^i from i=0 to i=log2dim
	//shortcut (4^(log2dim+1)-1)/3 = (2^(2*(log2dim+1))-1)/3
	//codewise ((1<<((log2dim+1)<<1))-1)/3
	blockCount = ((1 << ((log2dim + 1) << 1)) - 1) / 3;
	main_console->Print("we would have created %d blocks with this method\n", blockCount);

	//alloc
	block = (ls_block_t *)malloc(blockCount * sizeof(ls_block_t));
	if (!block) {
		main_console->Print("failed to allocate block data\n");
		return false;
	}
	//zero
	memset(block, 0, blockCount * sizeof(ls_block_t));

	//go at it recursive-style
	if (!genBlockData(0,0,0,width-1,height-1)) {
		main_console->Print("failed to generate block data\n");
		return false;
	}

	return true;
}

int LODMethodTrial2::getLevel(int index) {
	assert(index >= 0 && index < blockCount);
	if (!index) return 0;
	return getLevel(getParent(index)) + 1;
}

void LODMethodTrial2::renderBlock(int index) {

	static vector_t stepCoeff = {0, 0.06f, 0};
//	VectorSet(&stepCoeff, 1, 0.02f, 0);
//	VectorSet(&stepCoeff, 0, 0.06f, 0);

	int xmin, ymin, xmax, ymax;
	getBlockStats(index, &xmin, &ymin, &xmax, &ymax);

#if 0	//oldschool center distance - bad idea
	int cx = (xmin + xmax) >> 1;
	int cy = (ymin + ymax) >> 1;
	float cz = data[cx + cy * width].height;
	vector_t del = {
		(float)cx - main_camera->position.x,
		(float)cy - main_camera->position.y,
		cz - main_camera->position.z};
	float dist = (float)sqrt(del.x * del.x + del.y * del.y + del.z * del.z);
	float step = stepCoeff.x + stepCoeff.y * dist + stepCoeff.z * dist * dist;
	if (step < 1.f) {
#else
	static float maxDist = 1.f;
	static CVarw *dfjklsd = new CVarw(&maxDist, "maxDist");

	int cx = (xmin + xmax) >> 1;
	int cy = (ymin + ymax) >> 1;
	vector_t del = {
		(float)cx - main_camera->position.x,
		(float)cy - main_camera->position.y,
		data[cx + cy * width].height - main_camera->position.z};
	//attempt at screen metric
	vector_t *fwd = main_camera->matrix.vec;
	float d = (float)(xmax - xmin) / VectorLength(&del);//VectorDot(&del, fwd);
	if (d > maxDist) {
#endif
		int fc = getFirstChild(index);
		if (fc < blockCount) {	//if we have children then render them and return
			for (int c = fc; c < fc + 4; c++) {
				renderBlock(c);
			}
			return;
		}	//else render ourselves
	}

	float t = 0;
	float it = 1;

	ls_block_t *b = block + index;

	assert(!(b->indexCount % 3));

	glBegin(GL_TRIANGLES);

	for (int j = 0; j < b->indexCount; j++) {
		
		int k = b->indexData[j];
		assert(k >= 0 && k < b->vtxCount);

		int l0 = b->vtxData[k];
		assert(l0 >= 0 && l0 < width * height);

		int y0 = l0 / width;
		int x0 = l0 - (y0 * width);

		ls_vertex_t *d0 = data + l0;

		int l1 = b->vtxData[b->vtxLERPData[k]];
		assert(l1 >= 0 && l1 < width * height);

		int y1 = l1 / width;
		int x1 = l1 - (y1 * width);

		ls_vertex_t *d1 = data + l1;

		float x = it * (float)x0 + t * (float)x1;
		float y = it * (float)y0 + t * (float)y1;
		float z = it * d0->height + t * d1->height;

#ifndef LIGHTMAP_TEXTURE
		glColor3ubv((unsigned char *)&d0->color);	//no inteprolation for this quite yet
#endif
		glVertex3f(x, y, z);
	}

	glEnd();
}

void LODMethodTrial2::render(int width, int height, ls_vertex_t *base) {

#if 1
	renderBlock(0);
#else

	static int allowLevel = 0;
	static CVarw *dslf = new CVarw(&allowLevel, "allowLevel");

	//hack - render the leaves to make sure they work
//	float t = main_timer->GetTime() / 10.0;	//gettime slows us down
//	t -= (float)(int)t;
//	t = (1.f - (float)cos(t * M_PI)) * 0.5f;
//	t *= t * t * t;
	static float t = 0;
	if (main_keyboard->keystate[VkKeyScan(']')]) t += 0.01f;
	if (main_keyboard->keystate[VkKeyScan('[')]) t -= 0.01f;
	if (t < 0) {
		allowLevel++;
		if (allowLevel > getLevel(blockCount-1)) {
			allowLevel = getLevel(blockCount-1);
			t = 0;
		} else {
			t = 1;
		}
	}
	if (t > 1) {
		allowLevel--;
		if (allowLevel < 0) {
			allowLevel = 0;
			t = 1;
		} else {
			t = 0;
		}
	}
//	t += 0.001;
//	if (t > 1) t = 0;
	//TODO - compute 't' per-tile
	float it = 1.f - t;


	for (int i = 0; i < blockCount; i++) {
		if (allowLevel == -1) {
			if (!isLeaf(i)) continue;
		} else {
			if (getLevel(i) != allowLevel) continue;
		}

		ls_block_t *b = block + i;

		assert(!(b->indexCount % 3));

		glBegin(GL_TRIANGLES);

		for (int j = 0; j < b->indexCount; j++) {
			
			int k = b->indexData[j];
			assert(k >= 0 && k < b->vtxCount);

			int l0 = b->vtxData[k];
			assert(l0 >= 0 && l0 < width * height);

			int y0 = l0 / width;
			int x0 = l0 - (y0 * width);

			ls_vertex_t *d0 = base + l0;

			int l1 = b->vtxData[b->vtxLERPData[k]];
			assert(l1 >= 0 && l1 < width * height);

			int y1 = l1 / width;
			int x1 = l1 - (y1 * width);

			ls_vertex_t *d1 = base + l1;

			float x = it * (float)x0 + t * (float)x1;
			float y = it * (float)y0 + t * (float)y1;
			float z = it * d0->height + t * d1->height;

#ifndef LIGHTMAP_TEXTURE
			glColor3ubv((unsigned char *)&d0->color);	//no inteprolation for this quite yet
#endif
			glVertex3f(x, y, z);
		}

		glEnd();

	}

#endif

}

void LODMethodTrial2::shutdown(int width, int height, ls_vertex_t *data) {
	if (block) {
		for (int i = 0; i < blockCount; i++) {
			if (block[i].indexData) free(block[i].indexData);
			if (block[i].vtxData) free(block[i].vtxData);
			if (block[i].vtxLERPData) free(block[i].vtxLERPData);
		}
		free(block);
		block = NULL;
	}
	blockCount = 0;
}

const char *LODMethodTrial2::desc(void) { return "trial 2 - GameDeveloper"; }

LODMethodTrial3::LODMethodTrial3() {

	VectorSet(&stepCoeff, 1, 0.02f, 0);
	maxQuadsPerBlock = 48;

	new CVarw(&stepCoeff, "t3_stepCoeff");
	new CVarw(&maxQuadsPerBlock, "t3_maxQuadsPerBlock");

}

void LODMethodTrial3::render(int width, int height, ls_vertex_t *base) {

	int tileWidth = width - 1;
	int tileHeight = height - 1;

	int quadsPerBlock = (width + height) / 16;	//32;
	if (quadsPerBlock > maxQuadsPerBlock) {
		quadsPerBlock = maxQuadsPerBlock;
	}

	int blocksWide = tileWidth / quadsPerBlock;
	int blocksHigh = tileHeight / quadsPerBlock;

	for (int j = 0; j < blocksHigh; j++) {
		for (int i = 0; i < blocksWide; i++) {

			int startX = (i * tileWidth) / blocksWide;
			int startY = (j * tileHeight) / blocksHigh;
			int endX = ((i + 1) * tileWidth) / blocksWide;
			int endY = ((j + 1) * tileHeight) / blocksHigh;

			int cx = (startX + endX) >> 1;
			int cy = (startY + endY) >> 1;
			float cz = base[cx + cy * width].height;
			vector_t del = {
				(float)cx - main_camera->position.x,
				(float)cy - main_camera->position.y,
				cz - main_camera->position.z};
			float dist = (float)sqrt(del.x * del.x + del.y * del.y + del.z * del.z);
			float step = stepCoeff.x + stepCoeff.y * dist + stepCoeff.z * dist * dist;
			if (step < 1.f) step = 1.f;
			
			if (startY < endY) {
				float y = (float)startY;
				bool doneY = false;

				int yInt = startY, nextYInt;
				float yFrac = 0, nextYFrac;
				float oneMinusYFrac = 1.f, oneMinusNextYFrac;

				do {
					float nextY = y + step;
					if (nextY >= (float)endY) {
						nextY = (float)endY;
						doneY = true;
					}

					nextYInt = (int)nextY;
					nextYFrac = nextY - (float)nextYInt;
					oneMinusNextYFrac = 1.f - nextYFrac;
					
					glBegin(GL_TRIANGLE_STRIP);

					if (startX < endX) {
						float x = (float)startX;
						int xInt = startX;
						float xFrac = 0;
						float oneMinusXFrac = 1;
						bool doneX = false;
						do {
							if (x >= (float)endX) {	//make sure we hit the last one.
								x = (float)endX;
								doneX = true;
							}
							xInt = (int)x;
							xFrac = x - (float)xInt;
							oneMinusXFrac = 1.f - xFrac;

							ls_vertex_t *data = base + xInt + nextYInt * width;
#ifndef LIGHTMAP_TEXTURE
							glColor3ubv((unsigned char *)&data->color);
#endif
							float z = data[0].height * oneMinusXFrac * oneMinusNextYFrac;
							if (xFrac) {
								z += data[1].height * xFrac * oneMinusNextYFrac;
								if (nextYFrac) {
									z += data[width+1].height * xFrac * nextYFrac;
								}
							}
							if (nextYFrac) {
								z += data[width].height * oneMinusXFrac * nextYFrac;
							}
							glVertex3f(x, nextY, z);

							data = base + xInt + yInt * width;
#ifndef LIGHTMAP_TEXTURE
							glColor3ubv((unsigned char *)&data->color);
#endif

							z = data[0].height * oneMinusXFrac * oneMinusYFrac;
							if (xFrac) {
								z += data[1].height * xFrac * oneMinusYFrac;
								if (yFrac) {
									z += data[width+1].height * xFrac * yFrac;
								}
							}
							if (yFrac) {
								z += data[width].height * oneMinusXFrac * yFrac;
							}
							glVertex3f(x, y, z);

							x += step;
						} while (!doneX);
					}
					glEnd();
					y = nextY;
					yInt = nextYInt;
					yFrac = nextYFrac;
					oneMinusYFrac = oneMinusNextYFrac;
				} while (!doneY);
			}
		}
	}
}

const char *LODMethodTrial3::desc(void) { return "trial 3 - fraction step"; }


LODMethodTrial4::LODMethodTrial4() {

	VectorSet(&stepCoeff, 0, 0.06f, 0);
	maxQuadsPerBlock = 256;//48;

	new CVarw(&stepCoeff, "t4_stepCoeff");
	new CVarw(&maxQuadsPerBlock, "t4_maxQuadsPerBlock");
	new CVarw(&blockVisCount, "t4_blockVisCount");

}

bool LODMethodTrial4::init(int width, int height, ls_vertex_t *data) {
	quadsPerBlock = (width + height) / 16;	//32;
	if (quadsPerBlock > maxQuadsPerBlock) {
		quadsPerBlock = maxQuadsPerBlock;
	}

	tileWidth = width - 1;
	tileHeight = height - 1;

	blocksWide = tileWidth / quadsPerBlock;
	blocksHigh = tileHeight / quadsPerBlock;

	block = (ls_block_t *)malloc(sizeof(ls_block_t) * blocksWide * blocksHigh);
	if (!block) return false;

	ls_block_t *b = block;
	for (int j = 0; j < blocksHigh; j++) {
		for (int i = 0; i < blocksWide; i++, b++) {

			b->startX = (i * tileWidth) / blocksWide;
			b->startY = (j * tileHeight) / blocksHigh;

			b->endX = ((i + 1) * tileWidth) / blocksWide;
			b->endY = ((j + 1) * tileHeight) / blocksHigh;

			b->minZ = fInf;
			b->maxZ = -fInf;
			for (int y = b->startY; y < b->endY; y++) {
				for (int x = b->startX; x < b->endX; x++) {
					ls_vertex_t *v = data + x + y*width;
					if (v->height < b->minZ) b->minZ = v->height;
					if (v->height > b->maxZ) b->maxZ = v->height;
				}
			}

			//we now have our block bounding volume.
		}
	}

	return true;
}

void LODMethodTrial4::shutdown(int width, int height, ls_vertex_t *data) {
	if (block) free(block);
	block = NULL;
}

void LODMethodTrial4::render(int width, int height, ls_vertex_t *base) {

	const float fLog2Inv = (float)(1.0 / log(2.0));

	box_t bbox;

	blockVisCount = 0;

	ls_block_t *b = block;
	for (int j = 0; j < blocksHigh; j++) {
		for (int i = 0; i < blocksWide; i++, b++) {

//			int startX = (i * tileWidth) / blocksWide;
//			int startY = (j * tileHeight) / blocksHigh;
//			int endX = ((i + 1) * tileWidth) / blocksWide;
//			int endY = ((j + 1) * tileHeight) / blocksHigh;

			bbox.min.x = (float)b->startX;
			bbox.min.y = (float)b->startY;
			bbox.min.z = b->minZ;
			bbox.max.x = (float)(b->endX - 1);
			bbox.max.y = (float)(b->endY - 1);
			bbox.max.z = b->maxZ;

			//bbox test to all camera planes
			for (int u = 0; u < 6; u++) {
				plane_t *p = main_camera->frustum_plane + u;

				//if it clips all the points of the bbox then its not visible
				//only test the nearest point 
				//dependant upon the normal...
				// ... if the normal x is positive then test the box x negative side...
				vector_t corner = {
					bbox.vec[p->normal.x < 0].x,
					bbox.vec[p->normal.y < 0].y,
					bbox.vec[p->normal.z < 0].z
				};

				if (PlaneVectorDistance(p, &corner) > 1.f) {	//epsilon
					break;
				}
			}
			//if we didn't make it thru them all then bail
			if (u < 6) continue;

			blockVisCount++;

			int cx = (b->startX + b->endX) >> 1;
			int cy = (b->startY + b->endY) >> 1;
			float cz = base[cx + cy * width].height;
			vector_t del = {
				(float)cx - main_camera->position.x,
				(float)cy - main_camera->position.y,
				cz - main_camera->position.z};
			float dist = (float)sqrt(del.x * del.x + del.y * del.y + del.z * del.z);
			float lodValue = stepCoeff.x + stepCoeff.y * dist + stepCoeff.z * dist * dist;
			if (lodValue < 1.f) lodValue = 1.f;
			float twoToStepFloat = (float)log(lodValue) * fLog2Inv;
			int twoToStep = (int)twoToStepFloat;
			float twoToStepFrac = twoToStepFloat - (float)twoToStep;
			int step = 1 << twoToStep;

			if (b->startY < b->endY) {
				int y = b->startY;
				bool doneY = false;
				bool yLerp = false;

				do {
					int nextY = y + step;
					bool nextYLerp = !yLerp;
					if (nextY >= b->endY) {
						nextY = b->endY;
						doneY = true;
					}
					int nextNextY = nextY + step;	
					bool nextNextYLerp = !nextYLerp;
					if (nextNextY >= b->endY) {					//clamp me.
						nextNextY = b->endY;
					}

					glBegin(GL_TRIANGLE_STRIP);

					if (b->startX < b->endX) {
						int x = b->startX;
						bool xLerp = false;
						int nextX = x + step;
						bool nextXLerp = !xLerp;
						bool doneX = false;
						do {
							if (x >= b->endX) {	//make sure we hit the last one.
								x = b->endX;
								doneX = true;
							}
							if (nextX >= b->endX) {	//clamp this beforehand if we have to
								nextX = b->endX;		//needed for interpolation
							}

#ifndef LIGHTMAP_TEXTURE
							glColor3ubv((unsigned char *)&base[x+y*width].color);
#endif
							float vx = (float)x;
							if (xLerp) {
								vx += (float)(nextX - vx) * twoToStepFrac;
							}
							float vy = (float)nextY;
							if (nextYLerp) {
								vy += (float)(nextNextY - vy) * twoToStepFrac;
							}
							float vz = base[x+nextY*width].height;
							if (xLerp && !nextYLerp) {
								vz += (base[nextX+nextY*width].height - vz) * twoToStepFrac;
							} else if (!xLerp && nextYLerp) {
								vz += (base[x+nextNextY*width].height - vz) * twoToStepFrac;
							} else if (xLerp && nextYLerp) {
								vz += (base[nextX+nextNextY*width].height - vz) * twoToStepFrac;
							}
							glVertex3f(vx, vy, vz);

#ifndef LIGHTMAP_TEXTURE
							glColor3ubv((unsigned char *)&base[x+nextY*width].color);
#endif
							//vx same as before
							vy = (float)y;
							if (yLerp) {
								vy += (float)(nextY - vy) * twoToStepFrac;
							}
							vz = base[x+y*width].height;
							if (xLerp && !yLerp) {
								vz += (base[nextX+y*width].height - vz) * twoToStepFrac;
							} else if (!xLerp && yLerp) {
								vz += (base[x+nextY*width].height - vz) * twoToStepFrac;
							} else if (xLerp && yLerp) {
								vz += (base[nextX+nextY*width].height - vz) * twoToStepFrac;
							}
							glVertex3f(vx, vy, vz);


							x = nextX;
							xLerp = nextXLerp;

							nextX += step;
							nextXLerp = !nextXLerp;

						} while (!doneX);
					}
					glEnd();
					y = nextY;
					yLerp = nextYLerp;
				} while (!doneY);
			}
		}
	}
}

const char *LODMethodTrial4::desc(void) { return "trial 4 - power of two fraction step"; }

