/*

    File: LODMethod.h

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#if !defined LODMETHOD_H
#define LODMETHOD_H

#include "landscape.h"

class LODMethod {
public:
	virtual bool init(int width, int height, ls_vertex_t *data);
	virtual void render(int width, int height, ls_vertex_t *data);
	virtual void shutdown(int width, int height, ls_vertex_t *data);
	virtual const char *desc(void);
};

class LODMethodBrute : public LODMethod {
public:
	virtual void render(int width, int height, ls_vertex_t *data);
	virtual const char *desc(void);
};


/**
 * the map is broken into blocks
 * each block is a fixed size
 * and contains static LOD meshes of tristrips
 * evenly spaced at further intervals for lower miplevels
 * LODs are hard-limit switched (no interpolation)
 * edges are not aligned or seamed, so we do get cracks.
 */
class LODMethodTrial1 : public LODMethod {
	vector_t stepCoeff;
	int maxQuadsPerBlock;
	bool forceAlignPowerOfTwo;
public:
	LODMethodTrial1();
//	virtual bool init(int width, int height, ls_vertex_t *data);
	virtual void render(int width, int height, ls_vertex_t *data);
	virtual const char *desc(void);
};

/**
 * Game Developer article algorithm - FAR FROM FINISHED
 */
class LODMethodTrial2 : public LODMethod {

	int width, height;
	ls_vertex_t *data;

	typedef struct {
//		int x0,y0,x1,y1;	//<- all this data can be computed by the block index.  dont store it.

		//indexCount/3 = triCount
		//indexData points to references in the vertex index table
		int indexCount;
		int *indexData;

		//vtxData contains ints equal to x+y*width to ref the vertex source
		//vtxLERPData references vtxData and points to where each vertex will interpolate
		int vtxCount;
		int *vtxData;
		int *vtxLERPData;
	} ls_block_t;

	int blockCount;
	ls_block_t *block;

	inline bool isLeaf(int index);
	inline int getParent(int childIndex);
	inline int getFirstChild(int parentIndex);
	inline int getChildRelative(int childIndex);
	inline void getBlockStats(int index, int *x0, int *y0, int *x1, int *y1);
	bool genBlockData(int index, int x0, int y0, int x1, int y1);

	int getLevel(int index);

	bool LODReduce(ls_block_t *b);

	void renderBlock(int index);

public:
	LODMethodTrial2();

	virtual bool init(int width, int height, ls_vertex_t *data);

	virtual void render(int width, int height, ls_vertex_t *data);

	virtual void shutdown(int width, int height, ls_vertex_t *data);

	virtual const char *desc(void);
};

/**
 * same as trial 1 but with fractional step calculations and
 * bilinear interpolation between fractions
 */
class LODMethodTrial3 : public LODMethod {
	vector_t stepCoeff;
	int maxQuadsPerBlock;
public:
	LODMethodTrial3();
	virtual void render(int width, int height, ls_vertex_t *data);
	virtual const char *desc(void);
};

/**
 * same as trial 3 but with forced power of two LODs
 * this allows us to use single linear interpolation rather than bilinear
 */
class LODMethodTrial4 : public LODMethod {
	vector_t stepCoeff;

	int quadsPerBlock,maxQuadsPerBlock;

	int tileWidth, tileHeight;
	int blocksWide, blocksHigh;

	typedef struct {
		int startX, endX;
		int startY, endY;
		float minZ, maxZ;
	} ls_block_t;

	ls_block_t *block;	//sized blocksWide x blocksHigh

	int blockVisCount;

public:
	LODMethodTrial4();

	virtual bool init(int width, int height, ls_vertex_t *data);
	virtual void shutdown(int width, int height, ls_vertex_t *data);

	virtual void render(int width, int height, ls_vertex_t *data);
	virtual const char *desc(void);
};
#endif