/*

    File: Landscape.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>
#include <gl/gl.h>
#include <math.h>

#include "../m_foundation/main.h"
#include "../m_foundation/fractal.h"
#include "../m_foundation/random.h"
#include "../m_math3d/vector.h"
#include "fractal.h"

#include "landscape.h"
#include "lodmethod.h"

#ifdef LIGHTMAP_TEXTURE
#include "../m_foundation/image.h"		//i've never used this class before...
#endif

#define numberof(x) (sizeof(x)/sizeof(x[0]))

enum {
	COMPILED_NO,
	COMPILED_YES,
	COMPILED_REQUEST,
};

static int isCompiled = COMPILED_NO;
static int compileID = 0;

static int usemultitex = 1;
static int lod_method_index = 1;
void CFunc_lod_method_index(int argc, char **argv);
#ifdef CALC_EDGE_COST
static int showCurv = 0;	//show the curvature
#endif

void CFunc_compile(int, char**) {
	switch (isCompiled) {
	case COMPILED_NO:
		isCompiled = COMPILED_REQUEST;
		compileID = glGenLists(1);
		break;

	case COMPILED_YES:
		glDeleteLists(compileID, 1);
		compileID = 0;
		isCompiled = COMPILED_NO;
		break;
	}
}

static CVarw *local_cvars[] = {
	new CVarw(&lod_method_index, "lod_method", CFunc_lod_method_index),
#ifdef CALC_EDGE_COST
	new CVarw(&showCurv,		"showCurv"),
#endif
	new CVarw(&usemultitex, "usemultitex"),
	new CVarw(0,				"compile",		CFunc_compile),
};


static ls_vertex_t *lsdata = NULL;

#define LS_WIDTH		257//1025//2048//512//128//64//1619//
#define LS_HEIGHT		LS_WIDTH
#define LS_ALT			((LS_HEIGHT + LS_WIDTH) / 16.f)
#define AMBIENT_LIGHT	0.3f

static Texture texture;

static LODMethod *method_array[] = {
	new LODMethodBrute(),
	new LODMethodTrial1(),
	new LODMethodTrial2(),
	new LODMethodTrial3(),
	new LODMethodTrial4(),
};

static LODMethod *lod_method = NULL;


//WARNING - dont attach this CFunc (or dont use it) until after 
//lsdata has been allocated
void CFunc_lod_method_index(int argc, char **argv) {
	LODMethod *old_lod_method = lod_method;

	if (lod_method_index < 0 || lod_method_index >= numberof(method_array)) {
		lod_method = NULL;
	} else {
		lod_method = method_array[lod_method_index];
		main_console->Print("switching to method:%s\n", lod_method->desc());
	}

	//if we are switching methods...
	if (old_lod_method != lod_method) {

		//shutdown the old method...
		if (old_lod_method) {
			old_lod_method->shutdown(LS_WIDTH, LS_HEIGHT, lsdata);
		}

		//init the new method...
		if (lod_method) {
			if (!lod_method->init(LS_WIDTH, LS_HEIGHT, lsdata)) {
				main_console->Print("failed to init method %s\n", lod_method->desc());
				lod_method = NULL;
			}
		}
	}
}

#ifdef CALC_EDGE_COST
//lower tris touch the y=0 edge along the x axis.
//upper tris touch the x=0 edge along the y axis
float computeCurvature(
	int *f, int fCount,		//negative f[] values denote 'use upper triangle with ~f[]'
	int *n, int nCount		//same goes for negative n's
) {

	float curvature = 0;

	for (int i = 0; i < fCount; i++) {
		vector_t n1;
		if (f[i] >= 0) {
			n1.x = lsdata[f[i]].height - lsdata[f[i]+1].height;
			n1.y = lsdata[f[i]+1].height - lsdata[f[i]+1+LS_WIDTH].height;
		} else {
			n1.x = lsdata[~f[i]+LS_WIDTH].height - lsdata[~f[i]+LS_WIDTH+1].height;
			n1.y = lsdata[~f[i]].height - lsdata[~f[i]+LS_WIDTH].height;
		}
		n1.z = 1;
		VectorUnit(&n1, &n1);

		float mincurv = 1;
		for (int j = 0; j < nCount; j++) {

			vector_t n2;
			if (n[j] >= 0) {
				n2.x = lsdata[n[j]].height - lsdata[n[j]+1].height;
				n2.y = lsdata[n[j]+1].height - lsdata[n[j]+1+LS_WIDTH].height;
			} else {
				n2.x = lsdata[~n[j]+LS_WIDTH].height - lsdata[~n[j]+LS_WIDTH+1].height;
				n2.y = lsdata[~n[j]].height - lsdata[~n[j]+LS_WIDTH].height;
			}
			n2.z = 1;
			VectorUnit(&n2, &n2);

			float dot = (1 - VectorDot(&n1,&n2)) * 0.5f;
			if (dot < mincurv) mincurv = dot;
		}
		if (mincurv > curvature) curvature = mincurv;
	}
	return curvature;
}
#endif

//this implementation uses the Image class under development
//#define USE_MF_IMAGE

#ifdef LIGHTMAP_TEXTURE
Texture lightmapTex;
#ifndef USE_MF_IMAGE
int lmWidth, lmHeight;
#endif
#endif
Texture colormapTex;

inline int fastRoundDownPowerOfTwo(int src) {
	src >>= 1;
	for (int dest=1; src; src >>= 1, dest <<= 1);
	return dest;
}

bool landscapeInit(void) {

	int x,y;

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_TEXTURE_2D);

	glClearColor(.3f, .4f, .5f, 1.f);

	lsdata = (ls_vertex_t *)malloc(LS_WIDTH * LS_HEIGHT * sizeof(ls_vertex_t));
	if (!lsdata) return false;

#if 0	//fractal generate texture
	//fractal generate 'ls_height'
	if (!Fractal_Create(
		&lsdata->height,
		0,
		LS_ALT,
		LS_WIDTH,
		LS_HEIGHT,
		sizeof(lsdata[0]),
		1))
	{
		return false;
	}
#else	//load from "c:\my documents\my pictures\ps_height_1k.bmp"
//	perlin_noise_t noise = {
//		//buffer specifications
//		&lsdata->height,//float	*ptr;
//		sizeof(lsdata[0]),//int		inc;		//increment between ptr's
//		LS_WIDTH,//int		width;
//		LS_HEIGHT,//int		height;
//		//perlin noise evaluation
//		0,//int		level_start;		//subdiv 2^[start,end]
//		10,//int		level_end;
//		0.5f,//double	persistance;
//		0xdeadbeef,//int		seed;
//	};
//	if (!PerlinNoise(&noise)) {
//		return false;
//	}
	const int seed = 0xdeadbeef;
	const int perlin_depth = 10;
	const float perlin_pers = 0.5f;
	const float perlin_amplitude = 1.f;
	const float perlin_base_freq = 8.f / (float)LS_WIDTH;
	if (!Perlin_Create(
		&lsdata->height,	//float *data,
		LS_WIDTH,			//int width,
		LS_HEIGHT,			//int height,
		sizeof(ls_vertex_t),	//int step_size,
		perlin_depth,
		perlin_pers,
		perlin_amplitude,
		perlin_base_freq,
		0, 0, seed))
	{
		main_console->Print("Perlin landscape creation failed\n");
		return false;
	}
	Normalize_Array(&lsdata->height, LS_WIDTH*LS_HEIGHT, sizeof(ls_vertex_t));
	Scale_Array(&lsdata->height, LS_WIDTH*LS_HEIGHT, sizeof(ls_vertex_t), LS_ALT);
#endif
#if 0
	Image lsImage;
	const char *lsfilename = "c:/my documents/my pictures/ps_height_1k.bmp";
	if (!lsImage.Load(lsfilename)) {
		main_console->Print("failed to load the image %s\n", lsfilename);
	} else {
		//WARNING - this is a mess 'cuz we need LS_WIDTH and LS_HEIGHT to match this exactly. sorta
		unsigned char *data = (unsigned char *)lsImage.image;
		for (int y = 0; y < LS_HEIGHT; y++) {
			for (int x = 0; x < LS_WIDTH; x++) {
				lsdata[x+y*LS_WIDTH].height = (float)data[
					((x*lsImage.width/LS_WIDTH) +
					(y*lsImage.width/LS_HEIGHT) * lsImage.width)*3] / 255.f * (LS_HEIGHT + LS_WIDTH) / 16.f;
			}
		}
		lsImage.Unload();
	}
#endif

//	texture[0].Create("../landscape/world/photo_grass.bmp");//landscape.bmp");
//	texture[1].Create("../landscape/world/photo_grass.bmp");//photo_grass.bmp");
//	texture[2].Create("../landscape/world/photo_grass.bmp");//photo_leaves.bmp");
	texture.Create("untitled-1.bmp");

	main_camera->position.x = LS_WIDTH * 0.5f;
	main_camera->position.y = LS_HEIGHT * 0.5f;
	main_camera->position.z = lsdata[(LS_WIDTH>>1) + (LS_HEIGHT>>1)*LS_WIDTH].height + 1.f;

	ls_vertex_t *ptr;

	//gen curv

#ifdef CALC_EDGE_COST
	//gen curvature
	ptr = lsdata;
	for (y = 0; y < LS_HEIGHT; y++) {
		for (x = 0; x < LS_WIDTH; x++, ptr++) {

			int f[6], fCount;
			int n[2], nCount;

			//define first vertex faces

			fCount = 0;
			if (x < LS_WIDTH-1 && y < LS_HEIGHT-1) {
				f[fCount] = x + y * LS_WIDTH;		fCount++;
				f[fCount] = ~(x + y * LS_WIDTH);	fCount++;
			}

			if (x > 0 && y < LS_HEIGHT-1) {
				f[fCount] = ~(x-1 + y * LS_WIDTH);		fCount++;
			}
			if (x < LS_WIDTH-1 && y > 0) {
				f[fCount] = x + (y-1) * LS_WIDTH;		fCount++;
			}
			if (x > 0 && y > 0) {
				f[fCount] = (x-1) + (y-1) * LS_WIDTH;	fCount++;
				f[fCount] = ~((x-1) + (y-1) * LS_WIDTH);	fCount++;
			}

			ptr->cost_u = -1;	//invalid value
			ptr->cost_v = -1;	//invalid value
			ptr->cost_uv = -1;	//invalid value

			//calc cost u

			nCount = 0;
			if (x < LS_WIDTH-1 && y < LS_HEIGHT-1) {
				n[nCount] = x + y * LS_WIDTH;	nCount++;
			}
			if (x < LS_WIDTH-1 && y > 0) {
				n[nCount] = ~(x + (y-1) * LS_WIDTH);	nCount++;
			}
			ptr->cost_u = computeCurvature(f, fCount, n, nCount);

			//calc cost v

			nCount = 0;
			if (x < LS_WIDTH-1 && y < LS_HEIGHT-1) {
				n[nCount] = ~(x + y * LS_WIDTH);	nCount++;
			}
			if (x > 0 && y < LS_HEIGHT-1) {
				n[nCount] = (x-1) + y * LS_WIDTH;	nCount++;
			}
			ptr->cost_v = computeCurvature(f, fCount, n, nCount);

			//calc cost uv

			nCount = 0;
			if (x < LS_WIDTH-1 && y < LS_HEIGHT-1) {
				n[nCount] = x + y * LS_WIDTH;	nCount++;
				n[nCount] = ~(x + y * LS_WIDTH);	nCount++;
			}
			ptr->cost_uv = computeCurvature(f, fCount, n, nCount);
		}
	}
#endif

	//gen color

#ifdef LIGHTMAP_TEXTURE
#ifdef USE_MF_IMAGE
	Image lightmapImage;
	if (!lightmapImage.createEmpty(LS_WIDTH, LS_HEIGHT, GL_LUMINANCE, GL_UNSIGNED_BYTE)) {
		main_console->Print("failed to create empty lightmap\n");
		return true;
	}
	unsigned char *lmsrc = (unsigned char *)lightmapImage.image;
#else
	lmWidth = fastRoundDownPowerOfTwo(LS_WIDTH);
	lmHeight = fastRoundDownPowerOfTwo(LS_HEIGHT);
	//round down using GL_PROXY_TEXTURE_2D
	for (;lmWidth && lmHeight; ) {
		glTexImage2D(GL_PROXY_TEXTURE_2D, 0, 1, lmWidth, lmHeight, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, NULL);

		int tw;
		glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &tw);
		if (tw) break;

		if (lmWidth > lmHeight)	{
			lmWidth >>= 1;
		} else {
			lmHeight >>= 1;
		}
	}
	//FORCE
	//lmWidth = 256;
	//lmHeight = 256;
	unsigned char *lmsrc = (unsigned char *)malloc(lmWidth * lmHeight);
	if (!lmsrc) return false;
#endif
#endif
	unsigned char *cmsrc = (unsigned char *)malloc(lmWidth * lmHeight * 3);
	if (!cmsrc) return false;

	vector_t light = {1,0,1};
	VectorUnit(&light, &light);
	vector_t n;
	ptr = lsdata;
	for (y = 0; y < LS_HEIGHT; y++) {
		for (x = 0; x < LS_WIDTH; x++, ptr++) {
			n.x = ptr->height - lsdata[((x+1)%LS_HEIGHT) + y*LS_WIDTH].height;
			n.y = ptr->height - lsdata[x + ((y+1)%LS_HEIGHT)*LS_WIDTH].height;
			n.z = 1.f / (float)sqrt(n.x*n.x + n.y*n.y + 1);
			n.x *= n.z;
			n.y *= n.z;
			//now dx,dy,dz are the normal vector components at this tile
			float dot = VectorDot(&light, &n);
			if (dot < AMBIENT_LIGHT) dot = AMBIENT_LIGHT;
			unsigned char c = (unsigned char)(255.f * dot);
#ifdef LIGHTMAP_TEXTURE
#ifdef USE_MF_IMAGE
			*lmsrc = c; lmsrc++;
			*lmsrc = c; lmsrc++;
			*lmsrc = c; lmsrc++;
#else
			lmsrc[ (x*lmWidth/LS_WIDTH) + (y*lmHeight/LS_HEIGHT)*lmWidth] = c;
#endif
#else
			ptr->color = c | (c << 8) | (c << 16) | (c << 24);
#endif

#ifdef COLORMAP_USE_COST
			float f = (ptr->cost_u + ptr->cost_v + ptr->cost_uv) / 3.f;
			cmsrc[ 0+3*((x*lmWidth/LS_WIDTH) + (y*lmHeight/LS_HEIGHT)*lmWidth)] = 0;
			cmsrc[ 1+3*((x*lmWidth/LS_WIDTH) + (y*lmHeight/LS_HEIGHT)*lmWidth)] = (unsigned char)(255.f * f);
			cmsrc[ 2+3*((x*lmWidth/LS_WIDTH) + (y*lmHeight/LS_HEIGHT)*lmWidth)] = 0;
#else
			static vector_t colors[] = {
				{0.8f,	0.6f,	0.7f},
				{0.6f,	0.5f,	0.5f},
				{0.2f,	0.4f,	0.3f},
				{0.5f,	0.7f,	0.4f},
				{0.3f,	0.6f,	0.3f},
				{0.5f,	0.6f,	0.5f},
				{0.7f,	0.75f,	0.8f},
				{1.f,	1.f,	1.f},
			};

			float height = (ptr->height / LS_ALT) * (float)(numberof(colors)-1);
			height += MCRand() * 0.5f;
			int colorIndex = (int)height;
			if (colorIndex < 0) colorIndex = 0;
			if (colorIndex >= numberof(colors)-1) colorIndex = numberof(colors)-2;
			float colorFrac = height - (float)colorIndex;
			if (colorFrac < 0) colorFrac = 0;
			if (colorFrac > 1) colorFrac = 1;
			vector_t v;
			VectorLerp(&colors[colorIndex], &colors[colorIndex+1], colorFrac, &v);
			cmsrc[ 0+3*((x*lmWidth/LS_WIDTH) + (y*lmHeight/LS_HEIGHT)*lmWidth)] = (unsigned char)(255.f * v.x);
			cmsrc[ 1+3*((x*lmWidth/LS_WIDTH) + (y*lmHeight/LS_HEIGHT)*lmWidth)] = (unsigned char)(255.f * v.y);
			cmsrc[ 2+3*((x*lmWidth/LS_WIDTH) + (y*lmHeight/LS_HEIGHT)*lmWidth)] = (unsigned char)(255.f * v.z);
#endif
		}
	}
#ifdef LIGHTMAP_TEXTURE
#ifdef USE_MF_IMAGE
	if (!lightmapImage.resample(1024, 1024)) {
		main_console->Print("failed to resample the lightmap!\n");
		return true;
		//or the max size that the card can handle
	}
	if (!lightmapImage.createTexture(&lightmapTex)) {
		main_console->Print("failed to create the lightmap texture!\n");
		return true;
		//or the max size that the card can handle
	}
	lightmapImage.Unload();
#else

//for (int j = 0; j < lmHeight; j++) {
//for (int i = 0; i < lmWidth; i++) {
//lmsrc[i+j*lmWidth] = ((i+j)&1)*255;
//}
//}

	if (!lightmapTex.Extract(lmWidth, lmHeight, GL_LUMINANCE, GL_UNSIGNED_BYTE, lmsrc, 1, lmWidth)) {
		main_console->Print("failed to extract the lightmap texture!\n");
		return true;
	}
	free(lmsrc);

//lightmapTex.SetFilter(GL_NEAREST, GL_NEAREST);
#endif
#endif

	if (!colormapTex.Extract(lmWidth, lmHeight, GL_RGB, GL_UNSIGNED_BYTE, cmsrc, 3, lmWidth*3)) {
		main_console->Print("failed to extract the colormap texture!\n");
		return true;
	}
	free(cmsrc);

//	CFunc_lod_method_index(0,0,0);	//start er up
	CVar_Exec("lod_method 4");	//my favorite so far

	return true;
}

void landscapeShutdown(void) {

	CVar_Exec("lod_method -1");	//shut er down

	for (int i = 0; i < numberof(method_array); i++) {
		delete method_array[i];
		method_array[i] = NULL;
	}

	if (lsdata) free(lsdata);
	lsdata = NULL;
}

void landscapeUpdate(void) {

	int texid[] = {
		texture.texture_id,
		texture.texture_id,
		texture.texture_id,
		texture.texture_id,
		texture.texture_id,
		texture.texture_id,
		colormapTex.texture_id,
		lightmapTex.texture_id
	};

	float texscale[][2] = {
		{16.f,	16.f},
		{4.f,	4.f},
		{1.f,	1.f},
		{0.25f,	0.25f},
		{0.0625f, 0.0625f},
		{0.015625f, 0.015625f},
		{1.f/(LS_WIDTH+2), 1.f/(LS_HEIGHT+2)},
		{1.f/(LS_WIDTH+2), 1.f/(LS_HEIGHT+2)},
	};

	glColor3f(1,1,1);

	if (lod_method) {
		for (int unit = 0; unit < numberof(texid); unit++) {

			bool thistimemtex = usemultitex && unit+1<numberof(texid);

			if (thistimemtex) {
				glActiveTextureARB(GL_TEXTURE0_ARB);
				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);//GL_DECAL);
			}
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, texid[unit]);
			glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
			glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
			glEnable(GL_TEXTURE_GEN_S);
			glEnable(GL_TEXTURE_GEN_T);

			glMatrixMode(GL_TEXTURE);
			glLoadIdentity();
			glScalef(texscale[unit][0], texscale[unit][1], 1.f);
			glMatrixMode(GL_MODELVIEW);

			if (thistimemtex) {
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, texid[unit+1]);
				glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
				glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
				glEnable(GL_TEXTURE_GEN_S);
				glEnable(GL_TEXTURE_GEN_T);

				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

				glMatrixMode(GL_TEXTURE);
				glLoadIdentity();
				glScalef(texscale[unit+1][0], texscale[unit+1][1], 1.f);
				//glTranslatef(-1.f / LS_WIDTH, -1.f / LS_HEIGHT, 0.f);
				glMatrixMode(GL_MODELVIEW);

				glActiveTextureARB(GL_TEXTURE0_ARB);
			}

			//int wx = (int)(main_camera->position.x * 2.f / (float)LS_WIDTH);
			//int wy = (int)(main_camera->position.y * 2.f / (float)LS_HEIGHT);
			//translate the landscape forward to meet the camera
			//push the camera->position
			//translate the camera->position inverse for calculations

			switch (isCompiled) {

			case COMPILED_YES:
				if (compileID) {
					glCallList(compileID);
					break;
				} else {
					main_console->Print("ERROR - got COMPILED_YES but found no list\n");
					isCompiled = COMPILED_NO;
				}
				//fall through

			case COMPILED_NO:
				lod_method->render(LS_WIDTH, LS_HEIGHT, lsdata);
				break;

			case COMPILED_REQUEST:
				glNewList(compileID, GL_COMPILE_AND_EXECUTE);
				lod_method->render(LS_WIDTH, LS_HEIGHT, lsdata);
				glEndList();
				isCompiled = COMPILED_YES;
				main_console->Print("setting state to COMPILED_YES...\n");
				break;
			}

			//pop the camera->position


			if (thistimemtex) {
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glDisable(GL_TEXTURE_GEN_S);
				glDisable(GL_TEXTURE_GEN_T);
				glDisable(GL_TEXTURE_2D);

				glMatrixMode(GL_TEXTURE);
				glLoadIdentity();
				glMatrixMode(GL_MODELVIEW);

				glActiveTextureARB(GL_TEXTURE0_ARB);

				unit++;
			}
			glDisable(GL_TEXTURE_GEN_S);
			glDisable(GL_TEXTURE_GEN_T);
			glDisable(GL_TEXTURE_2D);

			glMatrixMode(GL_TEXTURE);
			glLoadIdentity();
			glMatrixMode(GL_MODELVIEW);

			glDepthFunc(GL_EQUAL);
			glBlendFunc(GL_DST_COLOR, GL_ZERO);
			glEnable(GL_BLEND);
		}
	}
	glDepthFunc(GL_LESS);
	glDisable(GL_BLEND);

#ifdef CALC_EDGE_COST
	if (showCurv) {
		glBegin(GL_LINES);
		ls_vertex_t *ptr = lsdata;
		for (int j = 0; j < LS_HEIGHT; j++) {
			for (int i = 0; i < LS_WIDTH; i++, ptr++) {
				if (i < LS_WIDTH-1) {
					glColor3f(1, ptr->cost_u, 0);
					glVertex3f((float)i, (float)j, ptr->height);
					i++; ptr++;
					glColor3f(1, ptr->cost_u, 0);
					glVertex3f((float)i, (float)j, ptr->height);
					i--; ptr--;
				}
				if (j < LS_HEIGHT-1) {
					glColor3f(1, ptr->cost_v, 0);
					glVertex3f((float)i, (float)j, ptr->height);
					j++; ptr += LS_WIDTH;
					glColor3f(1, ptr->cost_v, 0);
					glVertex3f((float)i, (float)j, ptr->height);
					j--; ptr -= LS_WIDTH;
				}
				if (i < LS_WIDTH-1 && j < LS_HEIGHT-1) {
					glColor3f(1, ptr->cost_uv, 0);
					glVertex3f((float)i, (float)j, ptr->height);
					i++; j++; ptr += LS_WIDTH+1;
					glColor3f(1, ptr->cost_uv, 0);
					glVertex3f((float)i, (float)j, ptr->height);
					i--; j--; ptr -= LS_WIDTH+1;
				}
			}
		}
		glEnd();
	}
#endif
}
